@@ -16,47 +16,43 @@
 #include <QVBoxLayout>
 
 // KDE
-#include <klocalizedstring.h>
-#include <kmessagebox.h>
-#include <kopenwithdialog.h>
-#include <kpropertiesdialog.h>
+#include <KLocalizedString>
+#include <KMessageBox>
+#include <KOpenWithDialog>
+#include <KPropertiesDialog>
 
 // Local
 #include "kserviceselectdlg.h"
 #include "mimetypedata.h"
 
-KServiceListItem::KServiceListItem(const KService::Ptr &pService, int kind)
+KServiceListItem::KServiceListItem(const KService::Ptr &pService)
     : QListWidgetItem()
     , storageId(pService->storageId())
     , desktopPath(pService->entryPath())
 {
-    if (kind == KServiceListWidget::SERVICELIST_APPLICATIONS) {
-        setText(pService->name());
-    } else {
-        setText(i18n("%1 (%2)", pService->name(), pService->desktopEntryName()));
-    }
-
+    setText(pService->name());
     setIcon(QIcon::fromTheme(pService->icon()));
 
     if (!pService->isApplication()) {
-        localPath = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QStringLiteral("/kservices" QT_STRINGIFY(QT_VERSION_MAJOR))
-            + QLatin1Char('/') + desktopPath;
+        localPath = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation) + QStringLiteral("/kservices6/") + desktopPath;
     } else {
         localPath = pService->locateLocal();
     }
 }
 
-PluginListItem::PluginListItem(const KPluginMetaData &metaData)
+PluginListItem::PluginListItem(const KPluginMetaData &data)
     : QListWidgetItem()
+    , metaData(data)
 {
     setText(i18n("%1 (%2)", metaData.name(), metaData.pluginId()));
     setIcon(QIcon::fromTheme(metaData.iconName()));
 }
 
-KServiceListWidget::KServiceListWidget(int kind, QWidget *parent)
+KServiceListWidget::KServiceListWidget(Kind kind, QWidget *parent)
     : QGroupBox(kind == SERVICELIST_APPLICATIONS ? i18n("Application Preference Order") : i18n("Services Preference Order"), parent)
     , m_kind(kind)
     , m_mimeTypeData(nullptr)
+    , m_allowMultiApply(false)
 {
     QHBoxLayout *lay = new QHBoxLayout(this);
 
@@ -128,6 +124,13 @@ KServiceListWidget::KServiceListWidget(int kind, QWidget *parent)
     btnsLay->addWidget(servRemoveButton);
     servRemoveButton->setWhatsThis(i18n("Remove the selected application from the list."));
 
+    servApplyToButton = new QPushButton(i18n("Apply To..."), this);
+    servApplyToButton->setIcon(QIcon::fromTheme(QStringLiteral("edit-copy")));
+    servApplyToButton->setEnabled(false);
+    connect(servApplyToButton, &QAbstractButton::clicked, this, &KServiceListWidget::applyTo);
+    btnsLay->addWidget(servApplyToButton);
+    servApplyToButton->setWhatsThis(i18n("Apply the current preference order to other file types."));
+
     btnsLay->addStretch(1);
 }
 
@@ -145,22 +148,32 @@ void KServiceListWidget::setMimeTypeData(MimeTypeData *mimeTypeData)
     servicesLB->setEnabled(false);
 
     if (m_mimeTypeData) {
-        const QStringList services = (m_kind == SERVICELIST_APPLICATIONS) ? m_mimeTypeData->appServices() : m_mimeTypeData->embedServices();
-
-        if (services.isEmpty()) {
-            if (m_kind == SERVICELIST_APPLICATIONS) {
-                servicesLB->addItem(i18nc("No applications associated with this file type", "None"));
+        if (m_kind == SERVICELIST_APPLICATIONS) {
+            const QStringList services = m_mimeTypeData->appServices();
+            if (services.isEmpty()) {
+                if (m_kind == SERVICELIST_APPLICATIONS) {
+                    servicesLB->addItem(i18nc("No applications associated with this file type", "None"));
+                }
             } else {
-                servicesLB->addItem(i18nc("No components associated with this file type", "None"));
+                for (const QString &service : services) {
+                    if (KService::Ptr pService = KService::serviceByStorageId(service)) {
+                        servicesLB->addItem(new KServiceListItem(pService));
+                    }
+                }
+                servicesLB->setEnabled(true);
             }
         } else {
-            for (const QString &service : services) {
-                KService::Ptr pService = KService::serviceByStorageId(service);
-                if (pService) {
-                    servicesLB->addItem(new KServiceListItem(pService, m_kind));
+            const QStringList parts = m_mimeTypeData->embedParts();
+            if (parts.isEmpty()) {
+                servicesLB->addItem(new QListWidgetItem(i18nc("No components associated with this file type", "None"), nullptr, -1));
+            } else {
+                servicesLB->setEnabled(true);
+                for (const QString &partId : parts) {
+                    if (KPluginMetaData data(QStringLiteral("kf6/parts/") + partId); data.isValid()) {
+                        servicesLB->addItem(new PluginListItem(data));
+                    }
                 }
             }
-            servicesLB->setEnabled(true);
         }
     }
 
@@ -170,6 +183,18 @@ void KServiceListWidget::setMimeTypeData(MimeTypeData *mimeTypeData)
     if (servEditButton) {
         servEditButton->setEnabled(servicesLB->currentRow() > -1);
     }
+    if (servApplyToButton) {
+        servApplyToButton->setEnabled(m_allowMultiApply);
+    }
+}
+
+void KServiceListWidget::allowMultiApply(bool allow)
+{
+    m_allowMultiApply = allow;
+
+    if (m_mimeTypeData && servApplyToButton) {
+        servApplyToButton->setEnabled(m_allowMultiApply);
+    }
 }
 
 void KServiceListWidget::promoteService()
@@ -220,7 +245,6 @@ void KServiceListWidget::addService()
         return;
     }
 
-    KService::Ptr service;
     if (m_kind == SERVICELIST_APPLICATIONS) {
         KOpenWithDialog dlg(m_mimeTypeData->name(), QString(), this);
         dlg.setSaveNewApplications(true);
@@ -228,45 +252,39 @@ void KServiceListWidget::addService()
             return;
         }
 
-        service = dlg.service();
+        KService::Ptr service = dlg.service();
 
         Q_ASSERT(service);
         if (!service) {
             return; // Don't crash if KOpenWith wasn't able to create service.
         }
+        if (m_mimeTypeData->appServices().contains(service->storageId())) {
+            return;
+        }
+        servicesLB->insertItem(0, new KServiceListItem(service));
     } else {
-        KServiceSelectDlg dlg(m_mimeTypeData->name(), QString(), this);
+        KPartSelectDlg dlg(this);
         if (dlg.exec() != QDialog::Accepted) {
             return;
         }
-        service = dlg.service();
-        Q_ASSERT(service);
-        if (!service) {
+        const bool valid = dlg.chosenPart().isValid();
+        Q_ASSERT(valid);
+        if (m_mimeTypeData->embedParts().contains(dlg.chosenPart().pluginId())) {
             return;
         }
+        auto item = new PluginListItem(dlg.chosenPart());
+        servicesLB->insertItem(0, item);
     }
 
-    // Did the list simply show "None"?
-    const bool hadDummyEntry = (m_kind == SERVICELIST_APPLICATIONS) ? m_mimeTypeData->appServices().isEmpty() : m_mimeTypeData->embedServices().isEmpty();
-
-    if (hadDummyEntry) {
-        delete servicesLB->takeItem(0); // Remove the "None" item.
+    // We inserted a new service at the beginning, if we had a None entry before, it must
+    // be the second on in the element. Check the type to be sure of it
+    if (servicesLB->count() > 0 && servicesLB->item(1)->type() == -1) {
+        delete servicesLB->takeItem(1);
         servicesLB->setEnabled(true);
-    } else {
-        // check if it is a duplicate entry
-        for (int index = 0; index < servicesLB->count(); index++) {
-            if (static_cast<KServiceListItem *>(servicesLB->item(index))->desktopPath == service->entryPath()) {
-                // ##### shouldn't we make the existing entry the default one?
-                return;
-            }
-        }
     }
 
-    servicesLB->insertItem(0, new KServiceListItem(service, m_kind));
     servicesLB->setCurrentItem(nullptr);
-
     updatePreferredServices();
-
     Q_EMIT changed(true);
 }
 
@@ -335,7 +353,7 @@ void KServiceListWidget::editService()
 
     // ...and add it in the same place as the old one:
     if (addIt) {
-        servicesLB->insertItem(selected, new KServiceListItem(service, m_kind));
+        servicesLB->insertItem(selected, new KServiceListItem(service));
         servicesLB->setCurrentRow(selected);
     }
 
@@ -370,19 +388,29 @@ void KServiceListWidget::updatePreferredServices()
     }
     QStringList sl;
     unsigned int count = servicesLB->count();
-
     for (unsigned int i = 0; i < count; i++) {
-        KServiceListItem *sli = (KServiceListItem *)servicesLB->item(i);
-        sl.append(sli->storageId);
+        Q_ASSERT(servicesLB->item(i)->type() != -1);
+        if (m_kind == SERVICELIST_APPLICATIONS) {
+            auto sli = (KServiceListItem *)servicesLB->item(i);
+            sl.append(sli->storageId);
+        } else {
+            auto pluginItem = (PluginListItem *)servicesLB->item(i);
+            sl.append(pluginItem->metaData.pluginId());
+        }
     }
     sl.removeDuplicates();
     if (m_kind == SERVICELIST_APPLICATIONS) {
         m_mimeTypeData->setAppServices(sl);
     } else {
-        m_mimeTypeData->setEmbedServices(sl);
+        m_mimeTypeData->setEmbedParts(sl);
     }
 }
 
+void KServiceListWidget::applyTo()
+{
+    Q_EMIT multiApply(m_kind);
+}
+
 void KServiceListWidget::enableMoveButtons()
 {
     int idx = servicesLB->currentRow();
@@ -408,3 +436,5 @@ void KServiceListWidget::enableMoveButtons()
         servEditButton->setEnabled(m_kind == SERVICELIST_APPLICATIONS);
     }
 }
+
+#include "moc_kservicelistwidget.cpp"
